# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2012-2013, Andy Grover <agrover@redhat.com>
#
# Routines to export block devices over iscsi.

import contextlib
from rtslib import (Target, TPG, NodeACL, FabricModule, BlockStorageObject, RTSRoot,
                    NetworkPortal, LUN, MappedLUN, RTSLibError, RTSLibNotInCFS)
import lvm
from main import TargetdError
from utils import ignored


def get_vg_lv(pool_name):
    """
    Checks for the existence of a '/' in the pool name.  We are using this
    as an indicator that the vg & lv refer to a thin pool.
    """
    if '/' in pool_name:
        return pool_name.split('/')
    else:
        return pool_name, None


def pool_check(pool_name):
    """
    pool_name *cannot* be trusted, funcs taking a pool param must call
    this or vgopen() to ensure passed-in pool name is one targetd has
    been configured to use.
    """
    pool_to_check = get_vg_lv(pool_name)[0]

    if pool_to_check not in [get_vg_lv(x)[0] for x in pools]:
        raise TargetdError(-110, "Invalid pool")


@contextlib.contextmanager
def vgopen(pool_name):
    """
    Helper function to check/close vg for us.
    """
    global lib_calls
    pool_check(pool_name)
    with contextlib.closing(lvm.vgOpen(pool_name, "w")) as vg:
        yield vg

    # Clean library periodically
    lib_calls += 1
    if lib_calls > 50:
        try:
            # May not be present if using older library
            lvm.gc()
        except AttributeError:
            pass
        lib_calls = 0

pools = []
target_name = None
lib_calls = 0


#
# config_dict must include block_pools and target_name or we blow up
#
def initialize(config_dict):

    global pools
    pools = config_dict['block_pools']

    global target_name
    target_name = config_dict['target_name']

    # fail early if can't access any vg
    for pool in pools:
        test_vg = lvm.vgOpen(get_vg_lv(pool)[0])
        test_vg.close()

    return dict(
        vol_list=volumes,
        vol_create=create,
        vol_destroy=destroy,
        vol_copy=copy,
        export_list=export_list,
        export_create=export_create,
        export_destroy=export_destroy,
        initiator_set_auth=initiator_set_auth,
    )


def volumes(req, pool):
    output = []
    vg_name, lv_pool = get_vg_lv(pool)
    with vgopen(vg_name) as vg:
        for lv in vg.listLVs():
            attrib = lv.getAttr()
            if attrib[0] == 'V' or attrib[0] == '-':
                output.append(dict(name=lv.getName(), size=lv.getSize(),
                                   uuid=lv.getUuid()))
    return output


def create(req, pool, name, size):
    vg_name, lv_pool = get_vg_lv(pool)
    with vgopen(vg_name) as vg:
        if lv_pool:
            # Fall back to non-thinp if needed
            try:
                vg.createLvThin(lv_pool, name, int(size))
            except AttributeError:
                vg.createLvLinear(name, int(size))
        else:
            vg.createLvLinear(name, int(size))


def destroy(req, pool, name):
    with ignored(RTSLibNotInCFS):
        fm = FabricModule('iscsi')
        t = Target(fm, target_name, mode='lookup')
        tpg = TPG(t, 1, mode='lookup')

        so_name = "%s:%s" % (pool, name)
        if so_name in (lun.storage_object.name for lun in tpg.luns):
            raise TargetdError(-303, "Volume '%s' cannot be "
                                     "removed while exported" % name)

    with vgopen(get_vg_lv(pool)[0]) as vg:
        vg.lvFromName(name).remove()


def copy(req, pool, vol_orig, vol_new, timeout=10):
    """
    Create a new volume that is a copy of an existing one.
    Since 0.6, requires thinp support.
    """
    vg_name, thin_pool = get_vg_lv(pool)

    with vgopen(vg_name) as vg:
        if not thin_pool:
            raise RuntimeError("copy requires thin-provisioned volumes")

        try:
            vg.lvFromName(vol_orig).snapshot(vol_new)
        except AttributeError:
            raise NotImplementedError("liblvm lacks thin snap support")


def export_list(req):
    try:
        fm = FabricModule('iscsi')
        t = Target(fm, target_name, mode='lookup')
        tpg = TPG(t, 1, mode='lookup')
    except RTSLibNotInCFS:
        return []

    exports = []
    for na in tpg.node_acls:
        for mlun in na.mapped_luns:
            mlun_vg, mlun_name = \
                mlun.tpg_lun.storage_object.udev_path.split("/")[2:]
            with vgopen(get_vg_lv(mlun_vg)[0]) as vg:
                lv = vg.lvFromName(mlun_name)
                exports.append(
                    dict(initiator_wwn=na.node_wwn, lun=mlun.mapped_lun,
                         vol_name=mlun_name, pool=mlun_vg,
                         vol_uuid=lv.getUuid(), vol_size=lv.getSize()))
    return exports


def export_create(req, pool, vol, initiator_wwn, lun):
    # get wwn of volume so LIO can export as vpd83 info
    vg_name, thin_pool = get_vg_lv(pool)

    with vgopen(vg_name) as vg:
        vol_serial = vg.lvFromName(vol).getUuid()

    # only add new SO if it doesn't exist
    # so.name concats pool & vol names separated by ':'
    so_name = "%s:%s" % (vg_name, vol)
    try:
        so = BlockStorageObject(so_name)
    except RTSLibError:
        so = BlockStorageObject(so_name, dev="/dev/%s/%s" % (vg_name, vol))
        so.wwn = vol_serial

    # export useful scsi model if kernel > 3.8
    with ignored(RTSLibError):
        so.set_attribute("emulate_model_alias", '1')

    fm = FabricModule('iscsi')
    t = Target(fm, target_name)
    tpg = TPG(t, 1)
    tpg.enable = True
    tpg.set_attribute("authentication", '0')
    NetworkPortal(tpg, "0.0.0.0")
    na = NodeACL(tpg, initiator_wwn)

    # only add tpg lun if it doesn't exist
    for tmp_lun in tpg.luns:
        if tmp_lun.storage_object.name == so.name \
                and tmp_lun.storage_object.plugin == 'block':
            tpg_lun = tmp_lun
            break
    else:
        tpg_lun = LUN(tpg, storage_object=so)

    # only add mapped lun if it doesn't exist
    for tmp_mlun in tpg_lun.mapped_luns:
        if tmp_mlun.mapped_lun == lun:
            break
    else:
        MappedLUN(na, lun, tpg_lun)

    RTSRoot().save_to_file()


def export_destroy(req, pool, vol, initiator_wwn):
    pool_check(pool)
    fm = FabricModule('iscsi')
    t = Target(fm, target_name)
    tpg = TPG(t, 1)
    na = NodeACL(tpg, initiator_wwn)

    vg_name, thin_pool = get_vg_lv(pool)

    for mlun in na.mapped_luns:
        # all SOs are Block so we can access udev_path safely
        mlun_vg, mlun_name = \
            mlun.tpg_lun.storage_object.udev_path.split("/")[2:]

        if mlun_vg == vg_name and mlun_name == vol:
            tpg_lun = mlun.tpg_lun
            mlun.delete()
            # be tidy and delete unused tpg lun mappings?
            if not len(list(tpg_lun.mapped_luns)):
                so = tpg_lun.storage_object
                tpg_lun.delete()
                so.delete()
            break
    else:
        raise TargetdError(-151, "Volume '%s' not found in %s exports" %
                                 (vol, initiator_wwn))

    # Clean up tree if branch has no leaf
    if not len(list(na.mapped_luns)):
        na.delete()
        if not len(list(tpg.node_acls)):
            tpg.delete()
            if not len(list(t.tpgs)):
                t.delete()

    RTSRoot().save_to_file()


def initiator_set_auth(req, initiator_wwn, in_user, in_pass, out_user,
                       out_pass):
    fm = FabricModule('iscsi')
    t = Target(fm, target_name)
    tpg = TPG(t, 1)
    na = NodeACL(tpg, initiator_wwn)

    if not in_user or not in_pass:
        # rtslib treats '' as its NULL value for these
        in_user = in_pass = ''

    if not out_user or not out_pass:
        out_user = out_pass = ''

    na.chap_userid = in_user
    na.chap_password = in_pass

    na.chap_mutual_userid = out_user
    na.chap_mutual_password = out_pass

    RTSRoot().save_to_file()


def block_pools(req):
    results = []

    for pool in pools:
            with vgopen(get_vg_lv(pool)[0]) as vg:
                results.append(dict(name=pool, size=vg.getSize(),
                                    free_size=vg.getFreeSize(), type='block',
                                    uuid=vg.getUuid()))

    return results



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2012, Andy Grover <agrover@redhat.com>
#
# A server that exposes a network API for configuring
# sharable resources on the local machine, such as the LIO
# kernel target.

import os
import setproctitle
import json
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
import yaml
import itertools
import socket
import ssl
import traceback
import logging as log
from utils import TargetdError

default_config_path = "/etc/target/targetd.yaml"

default_config = dict(
    block_pools=['vg-targetd'],
    fs_pools=[],
    user="admin",
    log_level='info',
    # security: no default password
    target_name="iqn.2003-01.org.linux-iscsi.%s:targetd" %
                socket.gethostname(),
    ssl=False,
    ssl_cert="/etc/target/targetd_cert.pem",
    ssl_key="/etc/target/targetd_key.pem",
)

config = {}

# Will be added to by fs/block.initialize()
mapping = dict()


class TargetHandler(BaseHTTPRequestHandler):

    def log_request(self, code='-', size='-'):
        # override base class - don't log good requests
        pass

    def do_POST(self):

        rpcdata = ""
        error = None

        # get basic auth string, strip "Basic "
        try:
            auth64 = self.headers.getheader("Authorization")[6:]
            in_user, in_pass = auth64.decode('base64').split(":")
        except:
            self.send_error(400)
            return

        if in_user != config['user'] or in_pass != config['password']:
            self.send_error(401)
            return

        if not self.path == "/targetrpc":
            self.send_error(404)
            return

        try:
            error = (-1, "jsonrpc error")
            self.id = None
            try:
                content_len = int(self.headers.getheader('content-length'))
                req = json.loads(self.rfile.read(content_len))
            except ValueError:
                # see http://www.jsonrpc.org/specification for errcodes
                error = (-32700, "parse error")
                raise

            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()

            try:
                version = req['jsonrpc']
                if version != "2.0":
                    raise ValueError
                method = req['method']
                self.id = int(req['id'])
                params = req.get('params', None)
            except (KeyError, ValueError):
                error = (-32600, "not a valid jsonrpc-2.0 request")
                raise

            try:
                if params:
                    result = mapping[method](self, **params)
                else:
                    result = mapping[method](self)
            except KeyError:
                error = (-32601, "method %s not found" % method)
                log.debug(traceback.format_exc())
                raise
            except TypeError:
                error = (-32602, "invalid method parameter(s)")
                log.debug(traceback.format_exc())
                raise
            except TargetdError, td:
                error = (td.error, td.msg)
                raise
            except Exception, e:
                error = (-1, "%s: %s" % (type(e).__name__, e))
                log.debug(traceback.format_exc())
                raise

            rpcdata = json.dumps(dict(result=result, id=self.id))

        except:
            log.debug('Error=%s, msg=%s' % error)
            rpcdata = json.dumps(
                dict(error=dict(code=error[0], message=error[1]), id=self.id))
        finally:
            self.wfile.write(rpcdata)


class HTTPService(HTTPServer, object):
    """
    Handle requests one at a time

    Note: The liblvm library we use is not thread safe, thus we need to
    serialize access to it.  It has locking for concurrent process usage, but
    we will keep things simple by serializing calls to it.
    """


class TLSHTTPService(HTTPService):
    """Also use TLS to encrypt the connection"""

    def finish_request(self, sock, addr):
        sockssl = ssl.wrap_socket(
            sock, server_side=True,
            keyfile=config["ssl_key"],
            certfile=config["ssl_cert"],
            ciphers="HIGH:-aNULL:-eNULL:-PSK",
            suppress_ragged_eofs=True)
        return self.RequestHandlerClass(sockssl, addr, self)


def load_config(config_path):
    global config

    if os.path.isfile(config_path):
        config = yaml.safe_load(open(config_path).read())
        if config is None:
            config = {}

    for key, value in default_config.iteritems():
        if key not in config:
            config[key] = value

    # compat: handle old single-pool config option
    if 'pool_name' in config:
        config['block_pools'] = [config['pool_name']]
        del config['pool_name']

    # uniquify pool lists
    config['block_pools'] = set(config['block_pools'])
    config['fs_pools'] = set(config['fs_pools'])

    if not config.get('password', None):
        log.critical("password not set in %s" % config_path)
        raise AttributeError

    # convert log level to int
    config['log_level'] = getattr(log, config['log_level'].upper(), log.INFO)
    log.basicConfig(level=config['log_level'])
    

def update_mapping():
    # wait until now so submodules can import 'main' safely
    import block
    import fs

    mapping.update(block.initialize(config))
    mapping.update(fs.initialize(config))

    # one method requires output from both modules
    def pool_list(req):
        return list(itertools.chain(block.block_pools(req), fs.fs_pools(req)))

    mapping['pool_list'] = pool_list


def main():
    server = None

    try:
        load_config(default_config_path)
    except AttributeError:
        return -1

    setproctitle.setproctitle("targetd")

    update_mapping()

    if config['ssl']:
        server_class = TLSHTTPService
        note = "(TLS yes)"
    else:
        server_class = HTTPService
        note = "(TLS no)"

    try:
        server = server_class(('', 18700), TargetHandler)
        log.info("started server %s", note)
        server.serve_forever()
    except KeyboardInterrupt:
        log.info("SIGINT received, shutting down")
        if server is not None:
            server.socket.close()
        return -1

    return 0



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2012, Andy Grover <agrover@redhat.com>
# Copyright 2013, Tony Asleson <tasleson@redhat.com>
#
# fs support using btrfs.

import os
import time
from nfs import Nfs, Export
from utils import invoke, TargetdError


# Notes:
#
# User can configure block pools (lvm volume groups) 1 to many or 0-many file
# system mount points to be used as pools.  At this time you have to specify
# a block pool for block operations and file system mount point pool for FS
# operations.  We could use files on a file system for block too and create
# file systems on top of lvm too, but that is TBD.
#
# We are using btrfs to provide all the cool fast FS features.  User supplies a
# btrfs mount point and we create a targetd_fs and targetd_ss subvolumes.  Each
# time the user creates a file system we are creating a subvolume under fs.
# Each time a FS clone is made we create the clone under fs.  For each snapshot
# (RO clone) we are creating a read only snapshot in
# <mount>/targetd_ss/<fsname>/<snapshot name>
#
# There may be better ways of utilizing btrfs.

import logging as log

fs_path = "targetd_fs"
ss_path = "targetd_ss"
fs_cmd = 'btrfs'


pools = []


def initialize(config_dict):

    global pools
    pools = config_dict['fs_pools']

    for pool in pools:
        # Make sure we have the appropriate subvolumes available
        try:
            create_sub_volume(os.path.join(pool, fs_path))
            create_sub_volume(os.path.join(pool, ss_path))
        except TargetdError, e:
            log.error('Unable to create required subvolumes')
            log.error(e.msg)
            raise

    return dict(
        fs_list=fs,
        fs_destroy=fs_destroy,
        fs_create=fs_create,
        fs_clone=fs_clone,
        ss_list=ss,
        fs_snapshot=fs_snapshot,
        fs_snapshot_delete=fs_snapshot_delete,
        nfs_export_auth_list=nfs_export_auth_list,
        nfs_export_list=nfs_export_list,
        nfs_export_add=nfs_export_add,
        nfs_export_remove=nfs_export_remove,
        )


def create_sub_volume(p):
    if not os.path.exists(p):
        invoke([fs_cmd, 'subvolume', 'create', p])


def split_stdout(out):
    """
    Split the text out as an array of text arrays.
    """
    rc = []
    for line in out.split('\n'):
        elem = line.split(' ')
        if len(elem) > 1:
            rc.append(elem)
    return rc


def fs_space_values(mount_point):
    """
    Return a tuple (total, free) from the specified path
    """
    st = os.statvfs(mount_point)
    free = (st.f_bavail * st.f_frsize)
    total = (st.f_blocks * st.f_frsize)
    return total, free


def pool_check(pool_name):
    """
    pool_name *cannot* be trusted, funcs taking a pool param must call
    this or to ensure passed-in pool name is one targetd has
    been configured to use.
    """
    if pool_name not in pools:
        raise TargetdError(-110, "Invalid filesystem pool")


def fs_create(req, pool_name, name, size_bytes):
    pool_check(pool_name)

    full_path = os.path.join(pool_name, fs_path, name)

    if not os.path.exists(full_path):
        invoke([fs_cmd, 'subvolume', 'create', full_path])
    else:
        raise TargetdError(-53, 'FS already exists')


def fs_snapshot(req, fs_uuid, dest_ss_name):
    fs = _get_fs_by_uuid(req, fs_uuid)

    if fs:
        source_path = os.path.join(fs['pool'], fs_path, fs['name'])
        dest_base = os.path.join(fs['pool'], ss_path, fs['name'])
        dest_path = os.path.join(dest_base, dest_ss_name)

        create_sub_volume(dest_base)

        if os.path.exists(dest_path):
            raise TargetdError(-53, "Snapshot already exists with that name")

        invoke([fs_cmd, 'subvolume', 'snapshot', '-r', source_path, dest_path])


def fs_snapshot_delete(req, fs_uuid, ss_uuid):
    fs_hash = _get_fs_by_uuid(req, fs_uuid)
    snapshot = _get_ss_by_uuid(req, fs_uuid, ss_uuid, fs_hash)
    path = os.path.join(fs_hash['pool'], ss_path, fs_hash['name'],
                        snapshot['name'])
    fs_subvolume_delete(path)


def fs_subvolume_delete(path):
    invoke([fs_cmd, 'subvolume', 'delete', path])


def fs_destroy(req, uuid):
    # Check to see if this file system has any read-only snapshots, if yes then
    # delete.  The API requires a FS to list its RO copies, we may want to
    # reconsider this decision.

    fs = _get_fs_by_uuid(req, uuid)

    base_snapshot_dir = os.path.join(fs['pool'], ss_path, fs['name'])

    snapshots = ss(req, uuid)
    for s in snapshots:
        fs_subvolume_delete(os.path.join(base_snapshot_dir, s['name']))

    if os.path.exists(base_snapshot_dir):
        fs_subvolume_delete(base_snapshot_dir)

    fs_subvolume_delete(os.path.join(fs['pool'], fs_path, fs['name']))


def fs_pools(req):
    results = []

    for pool in pools:
        total, free = fs_space_values(pool)
        results.append(dict(name=pool, size=total, free_size=free, type='fs'))

    return results


def _invoke_retries(command, throw_exception):
    # TODO take out this loop, used to handle bug in btrfs
    # ERROR: Failed to lookup path for root 0 - No such file or directory

    for i in range(0, 5):
        result, out, err = invoke(command, False)
        if result == 0:
            return result, out, err
        elif result == 19:
            time.sleep(1)
            continue
        else:
            raise TargetdError(-303, "Unexpected exit code %d" % result)

    raise TargetdError(-303, "Unable to execute command after "
                             "multiple retries %s" % (str(command)))


def _fs_hash():
    fs_list = {}

    for pool in pools:
        full_path = os.path.join(pool, fs_path)

        result, out, err = _invoke_retries(
            [fs_cmd, 'subvolume', 'list', '-ua', pool], False)

        data = split_stdout(out)
        if len(data):
            (total, free) = fs_space_values(full_path)
            for e in data:
                sub_vol = e[10]

                prefix = fs_path + os.path.sep

                if sub_vol[:len(prefix)] == prefix:
                    key = os.path.join(pool, sub_vol)
                    fs_list[key] = dict(name=sub_vol[len(prefix):],
                                        uuid=e[8],
                                        total_space=total,
                                        free_space=free,
                                        pool=pool,
                                        full_path=key)

    return fs_list


def fs(req):
    return _fs_hash().values()


def ss(req, fs_uuid, fs_cache=None):
    snapshots = []

    if fs_cache is None:
        fs_cache = _get_fs_by_uuid(req, fs_uuid)

    full_path = os.path.join(fs_cache['pool'], ss_path, fs_cache['name'])

    if os.path.exists(full_path):
        result, out, err = _invoke_retries([fs_cmd, 'subvolume', 'list', '-s',
                                            full_path], False)

        data = split_stdout(out)
        if len(data):
            for e in data:
                ts = "%s %s" % (e[10], e[11])
                time_epoch = int(time.mktime(
                    time.strptime(ts, '%Y-%m-%d %H:%M:%S')))
                st = dict(name=e[-1], uuid=e[-3], timestamp=time_epoch)
                snapshots.append(st)

    return snapshots


def _get_fs_by_uuid(req, fs_uuid):
    for f in fs(req):
        if f['uuid'] == fs_uuid:
            return f


def _get_ss_by_uuid(req, fs_uuid, ss_uuid, fs=None):
    if fs is None:
        fs = _get_fs_by_uuid(req, fs_uuid)

    for s in ss(req, fs_uuid, fs):
        if s['uuid'] == ss_uuid:
            return s


def fs_clone(req, fs_uuid, dest_fs_name, snapshot_id):
    fs = _get_fs_by_uuid(req, fs_uuid)

    if not fs:
        raise TargetdError(-104, "fs_uuid not found")

    if snapshot_id:
        snapshot = _get_ss_by_uuid(req, fs_uuid, snapshot_id)
        if not snapshot:
            raise TargetdError(-112, "snapshot not found")

        source = os.path.join(fs['pool'], ss_path, fs['name'], snapshot['name'])
        dest = os.path.join(fs['pool'], fs_path, dest_fs_name)
    else:
        source = os.path.join(fs['pool'], fs_path, fs['name'])
        dest = os.path.join(fs['pool'], fs_path, dest_fs_name)

    if os.path.exists(dest):
        raise TargetdError(-51, "Filesystem with that name exists")

    invoke([fs_cmd, 'subvolume', 'snapshot', source, dest])


def nfs_export_auth_list(req):
    return Nfs.security_options()


def nfs_export_list(req):
    rc = []
    exports = Nfs.exports()
    for e in exports:
        rc.append(dict(host=e.host, path=e.path, options=e.options_list()))
    return rc


def nfs_export_add(req, host, path, export_path, options):

    if export_path is not None:
        raise TargetdError(-401, "separate export path not supported at "
                                 "this time")
    bit_opt = 0
    key_opt = {}

    for o in options:
        if '=' in o:
            k, v = o.split('=')
            key_opt[k] = v
        else:
            bit_opt |= Export.bool_option[o]

    Nfs.export_add(host, path, bit_opt, key_opt)


def nfs_export_remove(req, host, path):
    found = False

    for e in Nfs.exports():
        if e.host == host and e.path == path:
            Nfs.export_remove(e)
            found = True

    if not found:
        raise TargetdError(-400, "NFS export to remove not found %s:%s",
                       (host, path))



#!/usr/bin/env python

from distutils.core import setup

setup(
    name = 'targetd',
    version = '0.7',
    description = 'Linux remote storage API daemon',
    license = 'GPLv3',
    maintainer = 'Andy Grover',
    maintainer_email = 'agrover@redhat.com',
    url = 'http://github.com/agrover/targetd',
    packages = ['targetd'],
    scripts = ['scripts/targetd'],
    )



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import hashlib
import os
import os.path
import shlex
from utils import invoke


def md5(t):
    h = hashlib.md5()
    h.update(t)
    return h.hexdigest()


def make_line_array(out):
    """
    Split the text out as an array of text strings
    """
    rc = []
    for line in out.split('\n'):
        if len(line) > 1:
            rc.append(line)
    return rc


class Export(object):

    SECURE = 0x00000001
    RW = 0x00000002
    RO = 0x00000004
    SYNC = 0x00000008
    ASYNC = 0x00000010
    NO_WDELAY = 0x00000020
    NOHIDE = 0x00000040
    CROSS_MNT = 0x00000080
    NO_SUBTREE_CHECK = 0x00000100
    INSECURE_LOCKS = 0x00000200
    ROOT_SQUASH = 0x00000400
    NO_ROOT_SQUASH = 0x00000800
    ALL_SQUASH = 0x00001000
    WDELAY = 0x00002000
    HIDE = 0x00004000
    INSECURE = 0x00008000

    bool_option = dict(secure=SECURE, rw=RW, ro=RO, sync=SYNC, async=ASYNC,
                       no_wdelay=NO_WDELAY, nohide=NOHIDE,
                       cross_mnt=CROSS_MNT, no_subtree_check=NO_SUBTREE_CHECK,
                       insecure_locks=INSECURE_LOCKS, root_squash=ROOT_SQUASH,
                       all_squash=ALL_SQUASH, wdelay=WDELAY, hide=HIDE,
                       insecure=INSECURE, no_root_squash=NO_ROOT_SQUASH)

    key_pair = dict(mountpoint=str, mp=str, fsid=None, refer=str, replicas=str,
                    anonuid=int, anongid=int)

    export_regex = '([\/a-zA-Z0-9\.-_]+)[\s]+(.+)\((.+)\)'
    octal_nums_regex = r"""\\([0-7][0-7][0-7])"""

    @staticmethod
    def _join(sep, *strings_to_join):
        rc = ''
        for s in strings_to_join:
            if len(s):
                if len(rc):
                    rc += sep
                rc += s
        return rc

    @staticmethod
    def _bc(int_type):
        """
        Bit count.

        returns the number of bits set.
        """
        count = 0
        while int_type:
            int_type &= int_type - 1
            count += 1
        return count

    @staticmethod
    def _validate_options(options):

        if Export._bc(((Export.RW | Export.RO) & options)) == 2:
            raise ValueError("Both RO & RW set")

        if Export._bc(((Export.INSECURE | Export.SECURE) & options)) == 2:
            raise ValueError("Both INSECURE & SECURE set")

        if Export._bc(((Export.SYNC | Export.ASYNC) & options)) == 2:
            raise ValueError("Both SYNC & ASYNC set")

        if Export._bc(((Export.HIDE | Export.NOHIDE) & options)) == 2:
            raise ValueError("Both HIDE & NOHIDE set")

        if Export._bc(((Export.WDELAY | Export.NO_WDELAY) & options)) == 2:
            raise ValueError("Both WDELAY & NO_WDELAY set")

        if Export._bc(
                ((Export.ROOT_SQUASH | Export.NO_ROOT_SQUASH) & options)) > 1:
            raise ValueError("Only one option of ROOT_SQUASH, NO_ROOT_SQUASH, "
                             "can be specified")

        return options

    @staticmethod
    def _validate_key_pairs(kp):
        if kp:
            if isinstance(kp, dict):
                for k, v in kp.items():
                    if k not in Export.key_pair:
                        raise ValueError('option %s not valid' % k)

                return kp
            else:
                raise ValueError('key_value_options domain is None or dict')
        else:
            return {}

    def __init__(self, host, path, bit_wise_options=0, key_value_options=None):

        if host == '<world>':
            self.host = '*'
        else:
            self.host = host
        self.path = path
        self.options = Export._validate_options(bit_wise_options)
        self.key_value_options = Export._validate_key_pairs(key_value_options)

    @staticmethod
    def parse_opt(options_string):
        bits = 0
        pairs = {}

        if len(options_string):
            options = options_string.split(',')
            for o in options:
                if '=' in o:
                    # We have a key=value
                    key, value = o.split('=')
                    pairs[key] = value
                else:
                    bits |= Export.bool_option[o]

        return bits, pairs

    @staticmethod
    def parse_export(tokens):
        rc = []

        try:
            global_options = ''
            options = ''
            path = ''

            if len(tokens) >= 1:
                path = tokens[0]

                if len(tokens) > 1:
                    for t in tokens[1:]:

                        #Handle global options
                        if t[0] == '-' and not global_options:
                            global_options = t[1:]
                            continue

                        # Check for a host or a host with an options group
                        if '(' and ')' in t:
                            if t[0] != '(':
                                host, options = t[:-1].split('(')
                            else:
                                host = '*'
                                options = t[1:-1]
                        else:
                            host = t

                        rc.append(Export(host, path,
                                         *Export.parse_opt(
                                             Export._join(
                                                 ',',
                                                 global_options,
                                                 options))))
                else:
                    rc.append(Export('*', path))

        except Exception as e:
            return None

        return rc

    @staticmethod
    def parse_exports_file(f):
        rc = []

        with open(f, "r") as e_f:
            for line in e_f:
                exp = Export.parse_export(
                    shlex.split(Export._chr_encode(line), '#'))
                if exp:
                    rc.extend(exp)

        return rc

    @staticmethod
    def parse_exportfs_output(export_text):
        rc = []
        pattern = re.compile(Export.export_regex)

        for m in re.finditer(pattern, export_text):
            rc.append(Export(m.group(2), m.group(1),
                             *Export.parse_opt(m.group(3))))
        return rc

    @staticmethod
    def _append(s, a):
        if len(s):
            s = s + "," + a
        else:
            s = a
        return s

    def options_list(self):
        rc = []
        for k, v in self.bool_option.items():
            if self.options & v:
                rc.append(k)

        for k, v in self.key_value_options.items():
            rc.append('%s=%s' % (k, v))

        return rc

    def options_string(self):
        return ','.join(self.options_list())

    @staticmethod
    def _double_quote_space(s):
        if ' ' in s:
            return '"%s"' % s
        return s

    def __repr__(self):
        return "%s %s(%s)" % (Export._double_quote_space(self.path).ljust(50),
                                self.host, self.options_string())

    def export_file_format(self):
        return "%s %s(%s)\n" % (Export._double_quote_space(self.path),
                                self.host, self.options_string())

    @staticmethod
    def _chr_encode(s):
        # Replace octal values
        p = re.compile(Export.octal_nums_regex)

        for m in re.finditer(p, s):
            s = s.replace('\\' + m.group(1), chr(int(m.group(1), 8)))

        return s

    def __eq__(self, other):
        return self.path == other.path and self.host == other.host


class Nfs(object):
    """
    Python module for configuring NFS exports
    """
    CMD = 'exportfs'
    EXPORT_FILE = 'targetd.exports'
    EXPORT_FS_CONFIG_DIR = '/etc/exports.d'
    MAIN_EXPORT_FILE = '/etc/exports'

    def __init__(self):
        pass

    @staticmethod
    def security_options():
        return "sys", "krb5", "krb5i", "krb5p"

    @staticmethod
    def _save_exports():
        # Remove existing export
        config_file = os.path.join(Nfs.EXPORT_FS_CONFIG_DIR, Nfs.EXPORT_FILE)
        try:
            os.remove(config_file)
        except OSError:
            pass

        # Get exports in /etc/exports
        user_exports = Export.parse_exports_file(Nfs.MAIN_EXPORT_FILE)

        # Recreate all existing exports
        with open(config_file, 'w') as ef:
            for e in Nfs.exports():
                if e not in user_exports:
                    ef.write(e.export_file_format())

    @staticmethod
    def exports():
        """
        Return list of exports
        """
        rc = []
        ec, out, error = invoke([Nfs.CMD, '-v'])
        rc = Export.parse_exportfs_output(out)
        return rc

    @staticmethod
    def export_add(host, path, bit_wise_options, key_value_options):
        """
        Adds a path as an NFS export
        """
        export = Export(host, path, bit_wise_options, key_value_options)
        options = export.options_string()

        cmd = [Nfs.CMD]

        if len(options):
            cmd.extend(['-o', options])

        cmd.extend(['%s:%s' % (host, path)])

        ec, out, err = invoke(cmd, False)
        if ec == 0:
            Nfs._save_exports()
            return None
        elif ec == 22:
            raise ValueError("Invalid option: %s" % err)
        else:
            raise RuntimeError('Unexpected exit code "%s" %s, out= %s' %
                               (str(cmd), str(ec),
                                str(out + ":" + err)))

    @staticmethod
    def export_remove(export):
        ec, out, err = invoke([Nfs.CMD, '-u', '%s:%s' %
                                              (export.host, export.path)])

        if ec == 0:
            Nfs._save_exports()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2013, Andy Grover <agrover@redhat.com>
#
# Utility functions.

from subprocess import Popen, PIPE
from contextlib import contextmanager

@contextmanager
def ignored(*exceptions):
    try:
        yield
    except exceptions:
        pass


class TargetdError(Exception):
    def __init__(self, error_code, message, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)
        self.error = error_code
        self.msg = message


def invoke(cmd, raise_exception=True):
    """
    Exec a command returning a tuple (exit code, stdout, stderr) and optionally
    throwing an exception on non-zero exit code.
    """
    c = Popen(cmd, stdout=PIPE, stderr=PIPE)
    out = c.communicate()

    if raise_exception:
        if c.returncode != 0:
            cmd_str = str(cmd)
            raise TargetdError(-303, 'Unexpected exit code "%s" %s, out= %s' %
                                     (cmd_str, str(c.returncode),
                                      str(out[0] + out[1])))

    return c.returncode, out[0], out[1]
